//PolyminoToOpenSCAD.c - Alain Brobecker,
//2019/01/06: In ffen2tex.c: started %POLY, correct polyminoes drawing
//2020/06/15: Polygon edge tracking almost the same as in ffen2tex.c
//2020/12/07: Use Read2DObjectLib.c
//             Added "]);" at the end of Polyhedron types 0 and 1
//2020/12/08: Succeeded in removing bugs for rounded edges
//2020/12/09: Now PolyminoSquareSize and PolyminoDepth are floats.
//2020/12/14: Removed bug with floats by changing PlacePointWithHG from define to function
//2020/12/21: Pieces made of multiple polyminoes (eg mazes) are now supported
//            PolyminoDepth changed to PolyminoZTop and PolyminoZBottom
//2020/12/28: ER set to PolyminoSquareSize/10
//2021/01/06: ER,HG and $fn at the beginning of the file
//            Ability to disable spheres and cylinders on top/bottom face

//!!! Polygon with holes are not supported !!!

//HG: HalfGap
//ER: EdgeRadius

#include <stdio.h>  //file functions, etc...
#include <stdlib.h> //exit(), malloc()

#define FALSE           0
#define TRUE            1
#define MaxWidth 99                   //Also for READ 2D OBJECTS
#define MaxHeight 99
#define MaxSize (MaxWidth*MaxHeight)
#define ES 0

//*************************
//****                 ****
//**** READ 2D OBJECTS ****
//****                 ****
//*************************

//variables for 2D object conversions
FILE *filep;
unsigned char *InFile;
int InFileLength;


#include "Read2DObjectLib.c"
//  #define Object2DMaxWidth 49
//  #define Object2DMaxHeight 49
//  char Object2D[Object2DMaxSize];
//  int Object2DWidth,Object2DHeight;
//  int NamedIntegerValue;
//  double NamedFloatValue;
//int ReadInteger(char *s) : stops when non digit is encountered
//double ReadFloat(char *s): stops when non digit is encountered
//int ReadNextObject2D(unsigned char *In,int InLength,int InPos,char *Out) {
//  Returns width in Object2DWidth, height in Object2DHeight, and 2D object in Object2D[]
//  Also returns the position after end2d line or -1 if no board was found
//int ReadNamedInteger(unsigned char *Name,int NameLength,unsigned char *In,int InLength,int InPos) {
//  Returns integer value in NamedIntegerValue
//  Also returns the position after "Name=" or -1 if this was not found
//int ReadNamedFloat(unsigned char *Name,int NameLength,unsigned char *In,int InLength,int InPos) {
//  Returns double value in NamedFloatValue
//  Also returns the position after "Name=" or -1 if this was not found
//int SearchNext(unsigned char *String,int StringLength,
//               unsigned char *Text,int TextLength,int TextPos) {

void PlacePointWithHG(double x,double y,double z,int VertexType,int GAP) {
  printf("[%g",x);
  if(GAP==TRUE) { if((VertexType&2)==0) { printf("-HG"); } else { printf("+HG"); } }
  printf(",%g",y);
  if(GAP==TRUE) { if((VertexType&4)==0) { printf("-HG"); } else { printf("+HG"); } }
  printf(",%g",z);
  if(GAP==TRUE) { if((VertexType&8)==0) { printf("-HG"); } else { printf("+HG"); } }
  printf("]");
}

void PlacePointWithHGandER(double x,double y,double z,int VertexType,int GAP,int ERX,int ERY,int ERZ) {
  printf("[%g",x);
  if(ERX==TRUE) { if((VertexType&2)==0) { printf("-ER"); } else { printf("+ER"); } }
  if(GAP==TRUE) { if((VertexType&2)==0) { printf("-HG"); } else { printf("+HG"); } }
  printf(",%g",y);
  if(ERY==TRUE) { if((VertexType&4)==0) { printf("-ER"); } else { printf("+ER"); } }
  if(GAP==TRUE) { if((VertexType&4)==0) { printf("-HG"); } else { printf("+HG"); } }
  printf(",%g",z);
  if(ERZ==TRUE) { if((VertexType&8)==0) { printf("-ER"); } else { printf("+ER"); } }
  if(GAP==TRUE) { if((VertexType&8)==0) { printf("-HG"); } else { printf("+HG"); } }
  printf("]");
}

char Doubled[2*MaxWidth][2*MaxHeight];

//no need to make boundary tests because we added ES around when doubling
void FillDoubledWithES(int x,int y) {
  Doubled[x][y]=ES;
  if(Doubled[x-1][y]!=ES) { FillDoubledWithES(x-1,y); }
  if(Doubled[x][y-1]!=ES) { FillDoubledWithES(x,y-1); }
  if(Doubled[x+1][y]!=ES) { FillDoubledWithES(x+1,y); }
  if(Doubled[x][y+1]!=ES) { FillDoubledWithES(x,y+1); }
}


//**************
//****      ****
//**** MAIN ****
//****      ****
//**************
int main(int argc,char *argv[]) {
  double PolyminoSquareSize,PolyminoZTop,PolyminoZBottom;
  int PolyminoType;
  int x,y,value,StartX,StartY,StartOfPolyminoFound;
  int Direction,NextDirection,NextDirectionNoModulo;
  int FilePos;
  double ConvertX[2*MaxWidth],ConvertY[2*MaxHeight];
  int NbPointsInPoly;
  int ConvertDirX[4]={1,0,-1,0}; //0=right, 1=down, 2=left, 3=up (clockwise)
  int ConvertDirY[4]={0,1,0,-1};
  int Index,VertexIndex,VI2,VI3,VI4,VerticesIndex[999];
  double PolyX[999],PolyY[999];
  char PolyVertexType[999],VertexType,VertexType2,VertexTypeBottom,VertexType2Bottom,GAP;
  int NbPoints;
  int NbFaces,NbSpheres,NbCylinders;

  if(argc!=2) { printf("usage: %s file.txt\n",argv[0]); exit(1); }


  //**** Load InFile ****
  filep=fopen(argv[1],"rb");
  if(filep==0) { printf("error: %s not found\n",argv[1]); exit(1); }
  fseek(filep,0,2); // 2 for EOF
  InFileLength=(int) ftell(filep);
  fseek(filep,0,0); // 0 for SOF
  InFile=(unsigned char *) malloc((unsigned int) InFileLength);
  fread(InFile,1,InFileLength,filep);
  fclose(filep);

  if(ReadNamedInteger("PolyminoType",sizeof("PolyminoType")-1,InFile,InFileLength,0)==-1) { printf("error: you must define PolyminoType\n"); exit(1); }
  PolyminoType=NamedIntegerValue;
  if(ReadNamedFloat("PolyminoSquareSize",sizeof("PolyminoSquareSize")-1,InFile,InFileLength,0)==-1) { printf("error: you must define PolyminoSquareSize\n"); exit(1); }
  PolyminoSquareSize=NamedFloatValue;
  if(ReadNamedFloat("PolyminoZTop",sizeof("PolyminoZTop")-1,InFile,InFileLength,0)==-1) { printf("error: you must define PolyminoZTop\n"); exit(1); }
  PolyminoZTop=NamedFloatValue;
  if(ReadNamedFloat("PolyminoZBottom",sizeof("PolyminoZBottom")-1,InFile,InFileLength,0)==-1) { printf("error: you must define PolyminoZBottom\n"); exit(1); }
  PolyminoZBottom=NamedFloatValue;
  if(PolyminoType>=2) { printf("ER=%g;\n",PolyminoSquareSize/10); }
  if(PolyminoType>=4) { printf("$fn=40;\n"); }
  if((PolyminoType%2)==0) { GAP=FALSE; } else { GAP=TRUE; printf("HG=0.05;\n"); }
  printf("//PolyminoType=%d\n",PolyminoType);
  printf("//PolyminoSquareSize=%g\n",PolyminoSquareSize);
  printf("//PolyminoZTop=%g\n",PolyminoZTop);
  printf("//PolyminoZBottom=%g\n",PolyminoZBottom);

  FilePos=ReadNextObject2D(InFile,InFileLength,0,Object2D);
  if(FilePos==-1) { printf("no start2d / end2d tags in file\n"); exit(1); }

  //copy Object2D[y*Object2DWidth+x] in Doubled[x][y] by adding borders and doubling horizontally and vertically
  //         ........
  //         ...bbbb.
  //  a..    ...bbbb.
  //  aab -> .aaaabb.
  //  .bb    .aaaabb.
  //         .aa.....
  //         .aa.....
  //         ........
  for(y=0;y<2*Object2DHeight+2;y++) { for(x=0;x<2*Object2DWidth+2;x++) { Doubled[x][y]=ES; } }
  for(y=0;y<Object2DHeight;y++) {
    for(x=0;x<Object2DWidth;x++) {
      value=Object2D[y*Object2DWidth+x];
      Doubled[1+2*x][1+2*y]=value;
      Doubled[1+2*x+1][1+2*y]=value;
      Doubled[1+2*x][1+2*y+1]=value;
      Doubled[1+2*x+1][1+2*y+1]=value;
    }
  }
  //Create ConvertX and ConvertY coords lookup tables
  for(x=0;x<Object2DWidth;x++) {
    ConvertX[x*2]=((double) x)*PolyminoSquareSize;
    ConvertX[x*2+1]=((double) (x+1))*PolyminoSquareSize;
  }
  for(y=0;y<Object2DHeight;y++) {
    ConvertY[y*2]=((double) (Object2DHeight-y))*PolyminoSquareSize;
    ConvertY[y*2+1]=((double) (Object2DHeight-y-1))*PolyminoSquareSize;
  }

  do {
    //search the start of the polygon, ie bottom left
    y=1;
    StartOfPolyminoFound=FALSE;
    while((y<2*Object2DHeight+2) && (StartOfPolyminoFound==FALSE)) {
      x=1;
      while((x<2*Object2DWidth+2) && (StartOfPolyminoFound==FALSE)) {
        value=Doubled[x][y];
        if(value!=ES) {
          StartOfPolyminoFound=TRUE;
          StartX=x;
          StartY=y;
        }
        x++;
      }//next x
      y++;
    }//next y
    if(StartOfPolyminoFound==TRUE) {

//for(y=1;y<1+2*Object2DHeight;y++) {
//  for(x=1;x<2*Object2DWidth;x++) {
//    if(Doubled[x][y]==ES) { printf("."); } else { printf("#"); }
//  }
//  printf("\n");
//}

      NbPointsInPoly=0;
      PolyX[NbPointsInPoly]=ConvertX[StartX-1];
      PolyY[NbPointsInPoly]=ConvertY[StartY-1];
      NbPointsInPoly++;

      Direction=0;
      x=StartX;
      y=StartY;
      //follow the polygon
      do{
        x+=ConvertDirX[Direction];
        y+=ConvertDirY[Direction];
        NextDirection=(Direction+3)&3; //try to turn left
        while(Doubled[x+ConvertDirX[NextDirection]][y+ConvertDirY[NextDirection]]!=value) {
          NextDirection=(NextDirection+1)&3;
        }
        if(NextDirection!=Direction) {
          PolyX[NbPointsInPoly]=ConvertX[x-1];
          PolyY[NbPointsInPoly]=ConvertY[y-1];

          VertexType=0;
          if((x-1)%2==0) { VertexType+=2; }
          if((y-1)%2!=0) { VertexType+=4; } //reversed Y
          //check if the direction change is inner or outer edge
          NextDirectionNoModulo=NextDirection;
          if(NextDirectionNoModulo<Direction) { NextDirectionNoModulo+=4; }
          if(NextDirectionNoModulo-Direction!=1) { VertexType+=1; }
          VertexType=VertexType+(NextDirection<<4);
          PolyVertexType[NbPointsInPoly]=VertexType;
          Direction=NextDirection;
          NbPointsInPoly++;
        }
      }while((x!=StartX) || (y!=StartY));
      NbPointsInPoly--;
      PolyVertexType[0]=PolyVertexType[NbPointsInPoly]; //first one was not processed before, but was at the end

      //remove this polymino
       FillDoubledWithES(StartX,StartY);

      //VertexType:
      //  bit 0 : 0=outer corner/1=inner corner
      //  bit 1 : x 0=-HG-ER  1=+HG+ER
      //  bit 2 : y-+
      //  bit 3 : z-+
      //  bit 4+: direction in [0;3], 0=right, 16=down, 32=left, 48=up

      NbPoints=0;
      if(PolyminoType<=1) {
        //***************************
        //**** STANDARD POLYMINO ****
        //***************************
        //**** points ****
        printf("polyhedron(\npoints=[");
        for(Index=0;Index<NbPointsInPoly;Index++) {
          VertexType=PolyVertexType[Index]+8;
          PlacePointWithHG(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexType,GAP);
          printf(",");
          VertexType-=8;
          PlacePointWithHG(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP);
          if(Index!=NbPointsInPoly-1) { printf(","); }
          NbPoints+=2;
        }//next Index
        //**** faces in anticlockwise order****
        printf("],\nfaces=[");
        NbFaces=2;
        //bottom face
        printf("[");
        for(Index=NbPointsInPoly-1;Index>=0;Index--) { printf("%d",Index*2); if(Index!=0) { printf(","); } }
        //top face
        printf("],[");
        for(Index=0;Index<NbPointsInPoly;Index++) { printf("%d",Index*2+1); if(Index!=NbPointsInPoly-1) { printf(","); } }
        printf("]");
        //side faces
        for(Index=0;Index<NbPointsInPoly;Index++) { printf(",[%d,%d,%d,%d]",Index*2+1,Index*2,((Index+1)*2)%(2*NbPointsInPoly),((Index+1)*2+1)%(2*NbPointsInPoly)); NbFaces+=1; }
        printf("]);\n");
      } else {
        //************************************************
        //**** POLYMINO WITH ANGULAR OR ROUNDED EDGES ****
        //************************************************
        //first create polyhedron with angular edges
        //**** points ****
        printf("polyhedron(\npoints=[");
        VertexIndex=0;
        for(Index=0;Index<NbPointsInPoly;Index++) {
          VerticesIndex[Index]=VertexIndex;
          VertexType=PolyVertexType[Index];
          VertexTypeBottom=VertexType+8;
          PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,TRUE,TRUE,FALSE);
          printf(",");
          PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,TRUE,TRUE,FALSE);
          printf(",");
          if((VertexType&1)==0) { //outer corner
            switch(VertexType&48) {
              case  0: case 32: //next direction was right or left
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,FALSE,TRUE,TRUE);
                printf(",");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,FALSE,TRUE,TRUE);
                printf(",");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,TRUE,FALSE,TRUE);
                printf(",");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,TRUE,FALSE,TRUE);
                if(Index!=NbPointsInPoly-1) { printf(","); }
                break;
              case 16: case 48: //next direction was down or up
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,TRUE,FALSE,TRUE);
                printf(",");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,TRUE,FALSE,TRUE);
                printf(",");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,FALSE,TRUE,TRUE);
                printf(",");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,FALSE,TRUE,TRUE);
                if(Index!=NbPointsInPoly-1) { printf(","); }
                break;
            }
            VertexIndex+=6;
            NbPoints+=6;
          } else { //inner corner
            PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,FALSE,FALSE,TRUE);
            printf(",");
            PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,FALSE,FALSE,TRUE);
            if(Index!=NbPointsInPoly-1) { printf(","); }
            VertexIndex+=4;
            NbPoints+=4;
          }
        }//next Index
        //**** faces in anticlockwise order****
        printf("],\nfaces=[");
        NbFaces=2;
        //bottom face
        printf("[");
        for(Index=NbPointsInPoly-1;Index>=0;Index--) { printf("%d",VerticesIndex[Index]); if(Index!=0) { printf(","); } }
        //top face
        printf("],[");
        for(Index=0;Index<NbPointsInPoly;Index++) { printf("%d",VerticesIndex[Index]+1); if(Index!=NbPointsInPoly-1) { printf(","); } }
        printf("]");
        //triangles and rectangles for outer corners
        for(Index=0;Index<NbPointsInPoly;Index++) {
          if((PolyVertexType[Index]&1)==0) { //outer corner?
            VertexIndex=VerticesIndex[Index];
            printf(",[%d,%d,%d]",VertexIndex+4,VertexIndex+2,VertexIndex);
            printf(",[%d,%d,%d]",VertexIndex+1,VertexIndex+3,VertexIndex+5);
            printf(",[%d,%d,%d,%d]",VertexIndex+2,VertexIndex+4,VertexIndex+5,VertexIndex+3);
            NbFaces+=3;
          }
        }
        //rectangles between two vertices
        for(Index=0;Index<NbPointsInPoly;Index++) {
          VertexIndex=VerticesIndex[Index];
          VI2=VertexIndex+2;
          if((PolyVertexType[Index]&1)==0) { VI2+=2; } //skip two points on this vertex if outer corner
          VI4=VerticesIndex[(Index+1)%NbPointsInPoly];
          VI3=VI4+2;
          printf(",[%d,%d,%d,%d]",VertexIndex,VI4,VI3,VI2);
          printf(",[%d,%d,%d,%d]",VI2+1,VI3+1,VI4+1,VertexIndex+1);
          printf(",[%d,%d,%d,%d]",VI2,VI3,VI3+1,VI2+1);
          NbFaces+=3;
        }//next Index
        printf("]);\n");
        if(PolyminoType>=4) {
          //***********************
          //**** ROUNDED EDGES ****
          //***********************
          //  4,5: ! polymino with rounded edges on top and mess on bottom
          //  6,7: polymino with rounded edges everywhere
          //  8,9: ! polymino with rounded edges on bottom and mess on top
          NbSpheres=0;
          NbCylinders=0;
          for(Index=0;Index<NbPointsInPoly;Index++) {
            VertexType=PolyVertexType[Index];
            VertexTypeBottom=VertexType+8;
            VertexType2=PolyVertexType[(Index+1)%NbPointsInPoly];
            VertexType2Bottom=VertexType2+8;
            if((VertexType&1)==0) { //outer corner, print a sphere
              //spheres and vertical cylinders
              VertexTypeBottom=VertexType+8;
              if(PolyminoType>5) {
                printf("translate(");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,TRUE,TRUE,TRUE);
                printf(") sphere(ER);");
                NbSpheres++;
              }
              if(PolyminoType<8) {
                printf("translate(");
                PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,TRUE,TRUE,TRUE);
                printf(") sphere(ER);\n");
                NbSpheres++;
              }
              printf("translate(");
              PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,TRUE,TRUE,TRUE);
              printf(") cylinder(%g-%g-2*ER",PolyminoZTop,PolyminoZBottom);
              if(GAP==TRUE) { printf("-2*HG"); }
              printf(",ER,ER);\n");
              NbCylinders++;
            }
            //horizontal cylinders in two directions
            switch(VertexType&48) {
              case  0: //next direction was right (draw Index to Index+1)
                if(PolyminoType>5) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([0,90,0]) cylinder(%g-%g",PolyX[(Index+1)%NbPointsInPoly],PolyX[Index]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                if(PolyminoType<8) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([0,90,0]) cylinder(%g-%g",PolyX[(Index+1)%NbPointsInPoly],PolyX[Index]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                break;
              case 32: //next direction was left (so draw Index+1 to Index)
                if(PolyminoType>5) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[(Index+1)%NbPointsInPoly],PolyY[(Index+1)%NbPointsInPoly],PolyminoZBottom,VertexType2Bottom,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([0,90,0]) cylinder(%g-%g",PolyX[Index],PolyX[(Index+1)%NbPointsInPoly]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                if(PolyminoType<8) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[(Index+1)%NbPointsInPoly],PolyY[(Index+1)%NbPointsInPoly],PolyminoZTop,VertexType2,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([0,90,0]) cylinder(%g-%g",PolyX[Index],PolyX[(Index+1)%NbPointsInPoly]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                break;
    //le long de y    translate([x2-r0,y1+r0,z2-r0]) rotate ([-90,0,0]) cylinder(y2-y1-2*r0,r0,r0);
              case 16: //next direction was down (so draw Index+1 to Index)
                if(PolyminoType>5) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[(Index+1)%NbPointsInPoly],PolyY[(Index+1)%NbPointsInPoly],PolyminoZBottom,VertexType2Bottom,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([-90,0,0]) cylinder(%g-%g",PolyY[Index],PolyY[(Index+1)%NbPointsInPoly]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                if(PolyminoType<8) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[(Index+1)%NbPointsInPoly],PolyY[(Index+1)%NbPointsInPoly],PolyminoZTop,VertexType2,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([-90,0,0]) cylinder(%g-%g",PolyY[Index],PolyY[(Index+1)%NbPointsInPoly]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                break;
              case 48: //next direction was up (draw Index to Index+1)
                if(PolyminoType>5) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZBottom,VertexTypeBottom,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([-90,0,0]) cylinder(%g-%g",PolyY[(Index+1)%NbPointsInPoly],PolyY[Index]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                if(PolyminoType<8) {
                  printf("translate(");
                  PlacePointWithHGandER(PolyX[Index],PolyY[Index],PolyminoZTop,VertexType,GAP,TRUE,TRUE,TRUE);
                  printf(") rotate([-90,0,0]) cylinder(%g-%g",PolyY[(Index+1)%NbPointsInPoly],PolyY[Index]);
                  if((VertexType&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  if((VertexType2&1)==0) { printf("-ER"); if(GAP==TRUE) { printf("-HG"); } } else { printf("+ER"); if(GAP==TRUE) { printf("+HG"); } }
                  printf(",ER,ER);\n");
                  NbCylinders++;
                }
                break;
            }
          }//next Index
          printf("//NbSpheres=%d\n//NbCylinders=%d\n",NbSpheres,NbCylinders);
        }
      }
      printf("//NbPoints=%d\n//NbFaces=%d\n",NbPoints,NbFaces);
    }
  } while(StartOfPolyminoFound==TRUE);

  exit(0);
}
