//Read2DObjectLib.c - Alain Brobecker
//2019/08/19: Now reads named value from file too
//2019/07/16: Reads 2d object
//2020/07/01: Made it more like a library
//2020/12/09: Changed name ReadNumber to ReadInteger, but also put ReadNumber as a define for compatibility
//            Added ReadFloat from ffen2tex.c and ReadNamedFloat
//            Modified ReadFloat

//#include "Read2DObjectLib.c"
//  #define Object2DMaxWidth 49
//  #define Object2DMaxHeight 49
//  char Object2D[Object2DMaxSize];
//  int Object2DWidth,Object2DHeight;
//  int NamedIntegerValue;
//  double NamedFloatValue;
//int ReadInteger(char *s) : stops when non digit is encountered
//double ReadFloat(char *s): stops when non digit is encountered
//int ReadNextObject2D(unsigned char *In,int InLength,int InPos,char *Out) {
//  Returns width in Object2DWidth, height in Object2DHeight, and 2D object in Object2D[]
//  Also returns the position after end2d line or -1 if no board was found
//int ReadNamedInteger(unsigned char *Name,int NameLength,unsigned char *In,int InLength,int InPos) {
//  Returns integer value in NamedIntegerValue
//  Also returns the position after "Name=" or -1 if this was not found
//int ReadNamedFloat(unsigned char *Name,int NameLength,unsigned char *In,int InLength,int InPos) {
//  Returns double value in NamedFloatValue
//  Also returns the position after "Name=" or -1 if this was not found
//int SearchNext(unsigned char *String,int StringLength,
//               unsigned char *Text,int TextLength,int TextPos) {

#define ReadNumber ReadInteger
#define Object2DMaxWidth 99
#define Object2DMaxHeight 99
#define Object2DMaxSize (Object2DMaxWidth*Object2DMaxHeight)

//*************************
//****                 ****
//**** READ 2D OBJECTS ****
//****                 ****
//*************************


//**** READINTEGER ****
//Version that reads a number as long as it is possible
int ReadInteger(char *s) {
  int pos,sign,number,ValidNumber;
  unsigned char c;
  pos=0;
  while(s[pos]==' ') { pos++; } //skip spaces
  number=0;
  sign=1;
  ValidNumber=FALSE;
  c=s[pos];
  if(c=='-') { sign=-1; pos++; c=s[pos]; }
  while((c>='0') && (c<='9')) {
    ValidNumber=TRUE;
    number=number*10+c-'0';
    pos++;
    c=s[pos];
  }
  if(ValidNumber==FALSE) { printf("Error: %c should be a digit",s[pos]); exit(1); }
  return(number*sign);
}

//**** READFLOAT ****
double ReadFloat(char *s) {
  int pos,number,sign,divisor,ValidNumber;
  unsigned char c;
  double f;
  pos=0;
  while(s[pos]==' ') { pos++; } //skip spaces
  number=0;
  sign=1;
  divisor=0;
  ValidNumber=FALSE;
  c=s[pos];
  if(c=='-') { sign=-1; pos++; c=s[pos]; }
  while(((c>='0') && (c<='9')) || (c=='.')) {
    ValidNumber=TRUE;
    if((c=='.') && (divisor==0)) {
      divisor=1;
    } else  if((c>='0') && (c<='9')) {
      number=number*10+c-'0';
      if(divisor!=0) { divisor=divisor*10; }
    } else {
      ValidNumber=FALSE;
    }
    pos++;
    c=s[pos];
  }
  if(ValidNumber==FALSE) { printf("Error: %c should be a digit or a dot",s[pos]); exit(1); }
  number=number*sign;
  if(divisor==0) {
    f=(double) number;
  } else {
    f=((double) (number))/((double) (divisor));
  }
  return(f);
}

//**** SEARCH NEXT ****
//SearchNext routine, brute force version
//To speed up this routine, copy beforehand the string to search at the
//end of the text (you must allocate a bit more memory) and remove the
//TextPos+StringLength<=TextLength condition in main loop. The routine
//will always find a string, once this is done, compare Pos+Length and
//return -1 if it is the one we added!
int SearchNext(unsigned char *String,int StringLength,
               unsigned char *Text,int TextLength,int TextPos) {
  int NbMatch=0;
  while((NbMatch<StringLength) && (TextPos+StringLength<=TextLength)) {
    if(Text[TextPos+NbMatch]==String[NbMatch]) {
      NbMatch++;
    } else {
      NbMatch=0;
      TextPos++;
    }
  }
  if(NbMatch==StringLength) {
    return(TextPos);
  } else {
    return(-1);
  }
}


//**** READ NEXT OBJECT 2D ****
//Returns width in Object2DWidth, height in Object2DHeight, and 2D object in Object2D[]
//Also returns the position after end2d line or -1 if no board was found
char Object2D[Object2DMaxSize];
int Object2DWidth,Object2DHeight;
int ReadNextObject2D(unsigned char *In,int InLength,int InPos,char *Out) {
  int StartLinePos,EndLinePos,BoardStartPos,BoardEndPos;
  char DefinedChar[256],ValueForChar[256];
  int pos,NbDefinedChars;
  int OutPos,CurrentWidth,NewOutLine,Object2DWidthDefined;
  unsigned char c;

  //search position of next line containing "start2d" and of line after it containing "end2d"
  StartLinePos=SearchNext((unsigned char *) "start2d",7,In,InLength,InPos);
  if(StartLinePos==-1) { return(-1); }
  EndLinePos=SearchNext((unsigned char *) "end2d",5,In,InLength,StartLinePos);
  if(EndLinePos==-1) { return(-1); }

  //Read definitions for the chars used in the next 2D board
  //They are after the "start2d" tag, in the form char=number
  for(c=0;c<255;c++) { DefinedChar[c]=FALSE; }
  pos=StartLinePos;
  NbDefinedChars=0;
  do {
    c=In[pos];
    pos++;
    if(c=='=') {
      //new definition
      c=In[pos-2];
      if(DefinedChar[c]==FALSE) { NbDefinedChars++; } //allow people to redefine chars
      DefinedChar[c]=TRUE;
      ValueForChar[c]=ReadInteger(In+pos);
    }
  } while((c!=10) && (c!=13)); //there might be a problem if someone writes "start2d end2d" on the same line

  if(NbDefinedChars==0) { printf("ReadNextObject2D Error: you must redefine at least one char after start2d\n"); exit(1); }

  //board starts after the CR/LF of line containing "start2d"
  BoardStartPos=StartLinePos+7;
  do{
    c=In[BoardStartPos];
    BoardStartPos++;
  } while((c!=10) && (c!=13));
  //board ends before the CR/LF that precedes "end2d"
  BoardEndPos=EndLinePos;
  do {
    c=In[BoardEndPos];
    BoardEndPos--;
  } while((c!=10) && (c!=13));

  //convert and compute Object2DWidth and Object2DHeight
  pos=BoardStartPos;
  OutPos=0;
  NewOutLine=TRUE;
  Object2DHeight=0;
  Object2DWidthDefined=FALSE;
  CurrentWidth=0;
  do{
    c=In[pos];
    if(DefinedChar[c]==TRUE) {
      Out[OutPos]=ValueForChar[c];
      OutPos++;
      if(NewOutLine==TRUE) {
        CurrentWidth=0;
        Object2DHeight++;
        if(Object2DHeight>Object2DMaxHeight) { printf("ReadNextObject2D Error: max height for objects is %d\n",Object2DMaxHeight); exit(1); }
        NewOutLine=FALSE;
      }
      CurrentWidth++;
    } else {
      //end of line reached?
      if((c==10) || (c==13)) {
        NewOutLine=TRUE;
        if(CurrentWidth!=0) {
          if(Object2DWidthDefined==FALSE) {
            Object2DWidth=CurrentWidth;
            if(Object2DWidth>Object2DMaxWidth) { printf("ReadNextObject2D Error: max width for objects is %d\n",Object2DMaxWidth); exit(1); }
            Object2DWidthDefined=TRUE;
          } else {
            if(Object2DWidth!=CurrentWidth) { printf("ReadNextObject2D Error: not the same width\n"); exit(1); }
          }
          CurrentWidth=0;
        }
      }
    }
    pos++;
  }while(pos<BoardEndPos+5);
  return(EndLinePos);
}

//**** READ NAMED INTEGER ****
//Returns integer value in NamedIntegerValue
//Also returns the position after "Name=" or -1 if this was not found
int NamedIntegerValue;
int ReadNamedInteger(unsigned char *Name,int NameLength,unsigned char *In,int InLength,int InPos) {
  int pos;
  pos=SearchNext(Name,NameLength,In,InLength,InPos)+NameLength; //skip the name
  if(pos==-1) { return(-1); }
  if(In[pos]!='=') { printf("error: %s must be followed by =value\n",Name); exit(1); }
  pos++;
  NamedIntegerValue=ReadInteger(In+pos);
  return(pos);
}

//**** READ NAMED FLOAT ****
//Returns integer value in NamedFloatValue
//Also returns the position after "Name=" or -1 if this was not found
double NamedFloatValue;
int ReadNamedFloat(unsigned char *Name,int NameLength,unsigned char *In,int InLength,int InPos) {
  int pos;
  pos=SearchNext(Name,NameLength,In,InLength,InPos)+NameLength; //skip the name
  if(pos==-1) { return(-1); }
  if(In[pos]!='=') { printf("error: %s must be followed by =value\n",Name); exit(1); }
  pos++;
  NamedFloatValue=ReadFloat(In+pos);
  return(pos);
}
